from typing import List

from pydantic import BaseModel
from pydantic.error_wrappers import ValidationError


class Validator:
    """
    Class responses for validation received object according to pydantic
    schema. After success validation gives possibility to get parsed
    instance os pydantic base class with all sent data.
    """
    def __init__(
            self,
            schema: BaseModel,
            response_data: dict,
            validation_key: str = None
    ) -> None:
        """
        Constructor of validator.
        :param schema: Pydantic schema. class BaseModel that will be applied
            to received data from BE for validation it.
        :param response_data: data received from BE
        :param validation_key: Validation key for getting concreate value from
            response data. Same parameter has class CamelConfig. Natively
            receiving this parameter in constructor has highly priority than
            parameter filled in the CamelConfig.
            For example: if you have dict like that:
            {"data": {"some": "data"}}, you can fill validation key with
            "data" value, so as a result, pydantic schema will be applied to
            this part of dict {"some": "data"}.
        """
        self.schema = schema
        self.response_data = response_data
        self.validation_key = validation_key

    def _iterator(
            self,
            searching_key,
            data_to_search=None
    ) -> [None, dict, list]:
        """
        Recursive method that try to detect part of object that should be
        validated according to received key. In case when key is absent,
        returns None.
        :param searching_key: string that equal to searching key in dict
        :param data_to_search: dict with data
        :return: return data according to searching key. If key is absent,
            returns None.
        """
        if data_to_search is None:
            data_to_search = self.response_data
        if isinstance(data_to_search, dict):
            for key in data_to_search:
                if key == searching_key:
                    return data_to_search.get(key)
                self._iterator(searching_key, data_to_search.get(key))
        return None

    def _data_searcher(self) -> dict:
        """
        According to set keys, try to get data from response data.
        If search key has been populated for more than 2 value, we try to
        get it without searching of around all data, if not, try to find needed
        data in dict by key.
        :return: Dict with concreate data for validation.
        """
        path = self.validation_key.split(':')
        if len(path) > 1:
            result = self.response_data
            for path_item in self.response_data:
                result = result.get(path_item)
        else:
            result = self._iterator(*path)
        return result

    def _validate(self, data_to_validate) -> List[BaseModel]:
        """
        Method check if data that he has is array or not, after it try
        to apply for each item pydantic schema. Instances of objects
        returns as array.
        :param data_to_validate: It could be dict or list.
        :return: list of instances of pydantic class BaseModel
        """
        result = []
        if isinstance(data_to_validate, list):
            for item in data_to_validate:
                result.append(self.schema.parse_obj(item))
        else:
            result.append(self.schema.parse_obj(data_to_validate))
        return result

    def fetch(self):
        """
        Method that applies validation.
        :return: List of instances of class BaseModel
        """
        if self.validation_key is not None:
            data_to_validate = self._data_searcher()
        else:
            data_to_validate = self.response_data
        try:
            initiated_objects = self._validate(data_to_validate)
            return initiated_objects
        except ValidationError as exception:
            raise AssertionError(
                f"\n\nException: {exception}"
                f"\nData passed to validator: {data_to_validate}"
                f"\nValidation schema: {self.schema.schema_json()}"
            ) from exception



def search_item(input_data, parameter):
    if isinstance(input_data, dict):
        for key, value in input_data.items():
            if key == parameter:
                yield value
            else:
                yield from search_item(value, parameter)
    elif isinstance(input_data, list):
        for item in input_data:
            yield from search_item(item, parameter)


def prepare_items(input_data, parameter):
    result = []
    params_iterator = search_item(input_data, parameter)
    while True:
        try:
            result.append(params_iterator.__next__())
        except StopIteration:
            break
    return result



import os


def search_key_processor(
        router_validation_key: str = None,
        response_validation_key: str = None
):
    env_key = os.getenv("pc_project_validation_key")
    if response_validation_key and response_validation_key != {}:
        return response_validation_key
    elif router_validation_key and router_validation_key != {}:
        return router_validation_key
    elif env_key and env_key != {}:
        return os.getenv("pc_project_validation_key")
    else:
        return None



from typing import List, Any

import requests
from json import JSONDecodeError

from pydantic import BaseModel
from pydantic.error_wrappers import ValidationError
from pycamel.src.modules.core.validator import Validator

from pycamel.src.utils.searcher import search_item, prepare_items
from pycamel.src.utils.search_key_processor import search_key_processor


class CamelResponse:

    def __init__(
            self,
            response: requests.Response,
            headers: dict,
            router_validation_key: str = None
    ) -> None:
        self.router_validation_key = router_validation_key
        self.response = response
        self.headers = headers
        self.validated_objects = []
        try:
            self.response_data = response.json()
        except JSONDecodeError:
            self.response_data = {}

    def assert_status_code(
            self,
            expected_status_code: List[int]
    ) -> 'CamelResponse':
        if not isinstance(expected_status_code, List):
            raise ValueError("Expected list of integers")
        else:
            assert self.response.status_code in expected_status_code, self
            return self

    def get_response_json(self) -> dict:
        return self.response_data

    def validate(
            self,
            schema: BaseModel,
            response_validation_key: str = None
    ) -> 'CamelResponse':
        validation_key = search_key_processor(
            self.router_validation_key, response_validation_key
        )
        try:
            self.validated_objects = Validator(
                schema, self.response_data, validation_key
            ).fetch()
            return self
        except ValidationError as validation_exception:
            raise AssertionError(
                "Received objects could not be mapped to schema"
            ) from validation_exception

    def assert_parameter(
                self,
                parameter: str,
                expected_value: Any
    ) -> 'CamelResponse':
        params_iterator = search_item(self.response_data, parameter)
        while True:
            try:
                item = params_iterator.__next__()
                assert item == expected_value, self
            except StopIteration:
                break
        return self

    def get_items_by_key(self, parameter: str) -> List:
        return prepare_items(self.response_data, parameter)

    def get_validated_objects(self) -> List:
        return self.validated_objects

# TODO Refactor error log. Split request and response data.
# TODO Add to error log request body if it was sent.
    def __str__(self) -> str:
        return f"\n\nRequest sent to: {self.response.url}\n" \
               f"Response status code: {self.response.status_code}\n" \
               f"Response data: {self.response_data}\n" \
               f"Headers: {self.headers}\n" \
               f"Request time took: {self.response.elapsed}"



from setuptools import setup, find_packages
import codecs
import os

"""
:authors: Yurii Abramenko
:license: Apache License, Version 2.0, see LICENSE file
:copyright: (c) 2022 SolveMeSolutions
"""

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.0.5'
DESCRIPTION = 'Backend automation framework. Automation is easy.'
LONG_DESCRIPTION = 'A package helps Automation QA creates automation' \
                   'project and setup tests without additional infrastructure '\
                   'and time wasting.' \
                   'All base structure has been written and prepared for you.'

# Setting up
setup(
    name="pycamel",
    version=VERSION,
    author="Yurii Abramenko",
    author_email="<yura.abramenko1@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    url='https://github.com/canyoupleasecreateanaccount/pycamel',
    license='Apache License, Version 2.0, see LICENSE file',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['pydantic', 'requests'],
    keywords=['python', 'automation', 'testing', 'tests', 'backend automation',
              'pytests', 'pydantic'],
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.0",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
        "Framework :: Pytest",
    ]
)
